package jarnal;

import java.awt.*;
import java.awt.font.*;
import java.awt.geom.*;
import java.util.*;
import java.io.*;
import java.awt.datatransfer.*;
import java.awt.image.BufferedImage;

class Jchar{
	private String aschar;
	public float size = 20.0f;
	private boolean bold = false;
	private boolean italic = false;
	private boolean underline = false;
	public String font = "Vera";
	public Color color = Color.black;
	private String cname = "black";
	private int hash;
	private boolean isBreak = false;
	private boolean isNull = false;

	public Jchar(){}
	public Jchar(String aschar, Jchar jc){
		if(aschar == "\n") isBreak = true;
		this.aschar = aschar;
		size = jc.size;
		bold = jc.bold;
		italic = jc.italic;
		underline = jc.underline;
		font = jc.font;
		color = jc.color;
		cname = jc.cname;
		setHash();	
	}
	public void clearStyle(){
		bold = false;
		italic = false;
		underline = false;
		color = Color.black;
		setHash();
	}
	public void setHash(){
		String svg = getSVG();
		hash = svg.hashCode();
	}
	public void setNull(){ isNull = true; isBreak = true;}
	public boolean isBreak(){return isBreak;}
	public boolean isNull(){return isNull;}
	public void setColor(Color color, String cname){ this.color = color; this.cname = cname; setHash();}
	public void setItalic(){ italic = true; setHash();}
	public void setBold(){ bold = true; setHash();}
	public void setUnderline(){ underline = true; setHash();}
	public void setSize(float size){ this.size = size; setHash();}
	public void setFont(String font){ this.font = font; setHash();}
	
	public String getChar(){ return aschar;}
	public int getStyle(){
		if(bold & italic) return Font.BOLD + Font.ITALIC;
		if(bold) return Font.BOLD;
		if(italic) return Font.ITALIC;
		return Font.PLAIN;
	}
	public String getSVG(){
		String svg = "";
		svg = svg + "font-family=\"" + font + "\" ";
		svg = svg + "font-size=\"" + (int) size + "\" ";	
		svg = svg + "fill=\"" + cname + "\" ";
		if(bold) svg = svg + "font-weight=\"bold\" ";
		if(italic) svg = svg + "font-style=\"italic\" ";
		if(underline) svg = svg + "text-decoration=\"underline\" ";
		return svg;
	}
	public String[] getHTML(){
		String html = "<FONT ";
		html = html + "COLOR=\"" + cname + "\" ";
		html = html + "FACE=\"" + font + "\" ";
		html = html + "POINT-SIZE=\"" + (int) size + "\">";
		String html2 = "</FONT>";	
		if(bold) {
			html = html + "<STRONG>";
			html2 = "</STRONG>" + html2;
		}
		if(italic) {
			html = html + "<I>";
			html2 = "</I>" + html2;
		}
		if(underline) {
			html = html + "<U>";
			html2 = "</U>" + html;
		}
		String ht[] = new String[2];
		ht[0] = html;
		ht[1] = html2;			
		return ht;
	}
	public String getDesc(){
		String d = "";
		d = font + " " + size + "pt " + cname;
		if(bold) d = d + " bold";
		if(italic) d = d + " italic";
		if(underline) d = d + " underline";
		return d;
	}
	public int getHash(){ return hash;}
}

public class Jpar{
	public static Graphics2D gg2 = (new BufferedImage(1, 1, BufferedImage.TYPE_INT_BGR)).createGraphics();
	public LinkedList chars = new LinkedList();
	private LinkedList jspans = new LinkedList();
	public int width;
	public int height = -1;
	private int ptr;
	private int dx;
	private int dy;
	private Jchar parms = new Jchar();
	private int sel = 0;
	private int selwidth = 0;
	private Jspan selSpan = null;

	public Jpar(){
		Jchar jc = new Jchar();
		jc.setNull();
		chars.add(jc);
	}
	public Jpar(String str){
		String a = getParm(str, "textLength=");	
		if(a == null) return;
		width = Integer.parseInt(a, 10);
		int pos = 0;
		String strstr;
		while(pos >=0){
			pos = str.indexOf("<tspan");
			if(pos >= 0){
				str = str.substring(pos);
				int last = str.indexOf("</tspan>");
				if(last < 0) return;
				strstr = str.substring(0, last);
				str = str.substring(last + 7);				
				makeSpan(strstr);
			}
		}
		Jchar jc = new Jchar();
		jc.setNull();
		chars.add(jc);
		sel = -1;
		if(gg2 == null) return;
		make();
		findSel();
	}

	public Jpar(LinkedList list, int w){
		chars = list;
		width = w;
		make();
	}

	public void setParms(Jchar parms){
		this.parms = parms;
	}

	public String getDesc(){ return parms.getDesc();}
	public Jchar getCurParms(){ return parms;}

	public void clearSel(){
		sel = -1;
		selSpan = null;
	}

	public void findSel(){
		if(sel == -1) {
			selSpan = null;
		}
		else{
			for(int i = 0; i <jspans.size(); i++){
				Jspan js = (Jspan) jspans.get(i);
				js.initSel();
			}
			int p = sel;
			if(selwidth < 0) p = sel + selwidth;
			for(int i = 0; i < jspans.size(); i++) {
				Jspan js = (Jspan) jspans.get(i);
				p = js.findSel(p);
				selSpan = js;
				if(p == -1) break;
			}
			if(selwidth != 0){
				p = sel + selwidth;
				if(selwidth < 0) p = sel;
				for(int i = 0; i < jspans.size(); i++) {
					Jspan js = (Jspan) jspans.get(i);
					p = js.findSel2(p);
					if(p == -1) break;
				}
			}
		}	
	}

	public void hitSel(Point2D.Double p, float s, boolean findWidth){
		int oldsel = sel;
		int dx = (int)(p.getX()/s);
		int dy = (int)(p.getY()/s);
		sel = 0;
		selSpan = null;
		for(int i = 0; i < jspans.size(); i++){
			Jspan js = (Jspan) jspans.get(i);
			int a = js.hitSel(dx, dy);
			sel = sel + Math.abs(a);
			if(a >= 0) break;
		}
		if(sel >= chars.size()) sel = chars.size() - 1;
		if(findWidth){
			selwidth = sel - oldsel;
			sel = oldsel;
		}
		else selwidth = 0;
		setParms();
	}

	public static Jchar makeParms(String str){
		Jchar jc = new Jchar();
		String a = getParm(str, "font-size=");
		if(a != null) jc.setSize((float) Integer.parseInt(a, 10));
		a = getParm(str, "font-family=");
		if(a != null) jc.setFont(a.trim());
		a = getParm(str, "fill=");
		if(a != null) {
			a = a.trim();
			jc.setColor(getColor(a), a);
		}
		a = getParm(str, "font-weight=");
		if(a != null) if(a.equals("bold")) jc.setBold();
		a = getParm(str, "font-style=");
		if(a != null) if(a.equals("italic")) jc.setItalic();
		a = getParm(str, "text-decoration=");
		if(a != null) if(a.equals("underline")) jc.setUnderline();
		return jc;
	}					

	public void makeSpan(String str){
		Jchar parms = new Jchar();
		if(str.equals("<tspan>") || str.equals("<tspan></tspan>")){
			chars.add(new Jchar("\n", parms));	
			return;
		}
		parms = makeParms(str);
		int last = str.indexOf(">");	
		str = str.substring(last + 1);
		last = str.indexOf("</tspan>");
		if(last > 0) str = str.substring(0, last);
		str = str.replaceAll("&lt;", "<");
		str = str.replaceAll("&gt;", ">");	
		for(int i = 0; i < str.length(); i++){
			chars.add(new Jchar(str.substring(i, i+1), parms));
		}
	}

	public static Color getColor(String color){
		Color c = Color.black;
		if(color.equals("blue")) c = Color.blue;
		if(color.equals("green")) c = Color.green;
		if(color.equals("dark gray")) c = Color.darkGray;
		if(color.equals("gray")) c = Color.gray;
		if(color.equals("light gray")) c = Color.lightGray;
		if(color.equals("magenta")) c = Color.magenta;
		if(color.equals("orange")) c = Color.orange;
		if(color.equals("pink")) c = Color.pink;
		if(color.equals("red")) c = Color.red;
		if(color.equals("white")) c = Color.white;
		if(color.equals("yellow")) c = Color.yellow;
		return c;
	}

	public static void setGraphics(Graphics2D gg2){
		Jpar.gg2 = gg2;
	}

	public static String getParm(String str, String parm){
		int pos = str.indexOf(parm);
		if(pos < 0) return null;
		str = str.substring(pos);
		pos = str.indexOf("\"");
		if (pos < 0) return null;
		str = str.substring(pos + 1);
		pos = str.indexOf("\"");
		if (pos < 0) return null;
		return str.substring(0, pos);
	}
	public void setStyle(Jchar jc, boolean bold, boolean italic, boolean underline, Float size, String font, String color){
		if(bold)jc.setBold();
		if(italic)jc.setItalic();
		if(underline)jc.setUnderline();
		if(size != null) jc.setSize(size.floatValue());
		if(font != null) jc.setFont(font);
		if(color != null) jc.setColor(getColor(color), color);			
	}
	public void setSelStyle(boolean bold, boolean italic, boolean underline, Float size, String font, String color){
		int start = sel;
		int stop = sel + selwidth;
		if(selwidth < 0){
			start = sel + selwidth;
			stop = sel;
		}
		Jchar jc;
		for(int i = start; i < stop; i++){
			jc = (Jchar) chars.get(i);
			setStyle(jc, bold, italic, underline, size, font, color);
		}
		setStyle(parms, bold, italic, underline, size, font, color);
		make();
		findSel();
	}
	public void setWidth(int w){ width = w;}
	public void setParms(){
		int p = sel;
		if(selwidth < 0) p = sel + selwidth;
		if(p > 0) {
			p = p - 1;
			Jchar jc = (Jchar) chars.get(p);
			if(jc == null) return;
			String s = jc.getChar();
			if(s != null){
				if(!(s.equals(" "))){
					parms = new Jchar(" ", jc);
					return;
				}
			}
		}
		p = sel;
		if(selwidth > 0) p = sel + selwidth;
		if(p >= chars.size()) p = chars.size() - 1;
		Jchar jc = (Jchar) chars.get(p);
		parms = new Jchar(" ", jc);
	}
	public void advSel(int adv, int extend){
		if(extend == 2){
			if(selwidth == 0) extend = 1;
		}
		if(extend == 1){
			selwidth = selwidth + adv;
			if(sel + selwidth < 0) selwidth = -sel;
			if(sel + selwidth >= chars.size()) selwidth = chars.size() - sel - 1;
		}
		if(extend == 0) {
			if(selwidth > 0) sel = sel + selwidth;
			selwidth = 0;
			sel = sel + adv;
			if(sel < 0) sel = 0;
			if(sel >= chars.size()) sel = chars.size() - 1;
		}
		setParms();
	}
	private void normalizeSel(){
		if(selwidth < 0){
			sel = sel + selwidth;
			selwidth = -selwidth;
		}
	}
	private LinkedList deleteSel(){
		LinkedList old = new LinkedList();
		if(selwidth == 0) return old;
		normalizeSel();
		for(int i = 0; i < selwidth; i++) old.add(chars.remove(sel));
		Jchar jc = (Jchar) old.getLast();
		if(jc.isNull()){
			chars.add(old.removeLast());
		}
		selwidth = 0;
		return old;
	}
	public void typeChar(String s){
		deleteSel();
		if(s.length() > 4){
			if(s.substring(0,5).equals("<text")){
				Jpar jp = new Jpar(s);
				parms = new Jchar(" ", (Jchar) jp.chars.getLast());
				for(int i = 0; i < jp.chars.size(); i++){
					chars.add(sel, jp.chars.get(i));
					sel++;
				}
				make();
				findSel();
				return;
			}
		}
		if(s.length() > 0) {
			for(int i = 0; i < s.length(); i++){
				Jchar jc = new Jchar(s.substring(i, i+1), parms);
				chars.add(sel, jc);
				sel++;
			}
		}
		make();
		findSel();
	}

	public String getText(){
		String txt = "";
		for(int i = 0; i < chars.size() - 1; i++){
			Jchar jc = (Jchar) chars.get(i);
			txt = txt + jc.getChar();
		}
		return txt;
	}
	public String getSVG(int x, int y){
		if(jspans.size() == 0){
			make();
			findSel();
		}
		String svg = "<text ";
		svg = svg + "x=\"" + x + "\" ";
		svg = svg + "y=\"" + y + "\" ";
		svg = svg + "textLength=\"" + width + "\" ";
		svg = svg + ">";
		for(int i = 0; i < jspans.size(); i++) {
			Jspan js = (Jspan) jspans.get(i);
			svg = svg + js.getSVG(x, y);
		}
		svg = svg + "</text>";	
		return svg;
	}
	public String getHTML(){
		if(jspans.size() == 0){
			make();
			findSel();
		}
		String html = "<html><head></head><body>";
		for(int i = 0; i < jspans.size(); i++) {
			Jspan js = (Jspan) jspans.get(i);
			html = html + js.getHTML();
		}
		html = html + "</body></html>";	
		return html;
	}
	public void clipSel(Toolkit toolkit){
		LinkedList clip = new LinkedList();
		int start = sel;
		int end = sel + selwidth;
		if(selwidth < 0) {
			start = sel + selwidth;
			end = sel;
		}
		for(int i = start; i < end; i++) clip.add(chars.get(i));
		Jpar jp = new Jpar(clip, width);
		jp.clip(toolkit);
	}
	public void clip(Toolkit toolkit){
		Clipboard clip = toolkit.getSystemClipboard();
		String copy = getSVG(0,0);
		String plain = 	getText();
		String html = getHTML();
		//StringSelection csel = new StringSelection(copy);
		JarnalSelection csel = new JarnalSelection(plain, copy, html);
		clip.setContents(csel, csel);
	}
	public void draw(int x, int y, float s, Graphics2D g2, boolean print){
		if(jspans.size() == 0){
			make();
			findSel();
		}
		if(selwidth != 0){
			for(int i = 0; i < jspans.size(); i++) {
				Jspan js = (Jspan) jspans.get(i);
				js.shadeSel(x, y, s, g2);
			}
		}
		for(int i = 0; i < jspans.size(); i++) {
			Jspan js = (Jspan) jspans.get(i);
			js.display(x, y, s, g2);
		}
		if((selwidth == 0) && (selSpan != null) && (!print))selSpan.displaySel(x, y, s, g2);
	}

	public FontMetrics setFont(Jchar jc){
		Font f = new Font(jc.font, jc.getStyle(), (int) (jc.size));
		gg2.setFont(f);
		return gg2.getFontMetrics();
	}

	public void make(){
		if(chars.size() == 0) return;
		ptr = 0;
		dx = 0;
		Jchar parms = (Jchar) chars.get(0);
		FontMetrics fm = setFont(parms);
		dy = fm.getAscent();
		jspans = new LinkedList();
		boolean ok = true;
		while(ok){
			Jspan js = findSpan();
			if(js == null) ok = false;
			else jspans.add(js);
		}
		height = dy;
	}
	private Jspan findSpan(){
		int adv[] = new int[400];
		int n = 0;
		int hash = 0;
		if(ptr >= chars.size()) return null;
		String txt = "";
		String wrd = "";
		boolean ok = true;
		Jchar parms = (Jchar) chars.get(ptr);
		Jchar jc = null;
		FontMetrics fm = setFont(parms);
		if(parms.isBreak()) {
			Jspan js = new Jspan(parms, dx, dy);
			dx = 0;
			dy = dy + fm.getAscent() + fm.getLeading();
			ptr++;
			if(parms.isNull())js.setNull();
			return js;
		}
		hash = parms.getHash();
		int olddx = dx;
		int olddy = dy;
		while(ok){
			if(ptr >= chars.size()){
				ok = false;
				txt = txt + wrd;
			}
			else{
				jc = (Jchar) chars.get(ptr);
				int nhash = jc.getHash();
				if((hash != nhash) || jc.isBreak()){
					ok = false;
					txt = txt + wrd;
				}
			}
			if(ok) {
				int m = fm.charWidth(jc.getChar().charAt(0));
				String c = jc.getChar();
				if(c.equals(" ")){
					txt = txt + wrd + " ";
					dx = dx + m;
					wrd = "";
					adv[n] = dx;
					n++;
					ptr++;
				}
				else {
					if(dx + m <= width) {
						wrd = wrd + jc.getChar();
						dx = dx + m;
						adv[n] = dx;
						n++;
						ptr++;
					}
					else{
						ok = false;
						dx = 0;
						dy = dy + fm.getAscent() + fm.getLeading();
						if(txt.length() > 0) ptr = ptr - wrd.length();
						else txt = txt + wrd;						
					}
				}
			}
		}
		int adv2[] = new int[n];
		for(int i = 0; i < txt.length(); i++) adv2[i]=adv[i];
		return new Jspan(parms, txt, adv2, olddx, olddy); 	
	}			
}

class Jspan{
	private Jchar parms;
	private String txt;
	private int adv[];
	private int dx;
	private int dy;
	private int sel = -1;
	private int sel2 = -2;
	private boolean isBreak = false;
	private boolean isNull = false;

	public Jspan(Jchar parms, String txt, int adv[], int dx, int dy){
		this.parms = parms;
		this.txt = txt;
		this.adv = adv;
		this.dx = dx;
		this.dy = dy;
	}

	public Jspan(Jchar parms, int dx, int dy){
		this.parms = parms;
		adv=new int[1];
		adv[0] = dx; 
		txt = " "; 
		isBreak = true;
		this.dx = dx;
		this.dy = dy;
	}

	public void setNull(){isNull = true;}

	public String getSVG(int x, int y){
		if(isNull) return "";
		if(isBreak) return "<tspan></tspan>";
		String rtxt = txt.replaceAll("<", "&lt;");
		rtxt = rtxt.replaceAll(">", "&gt;");
		String svg = "<tspan ";
		svg = svg + "x=\"" + (x + dx) + "\" ";
		svg = svg + "y=\"" + (y + dy) + "\" ";
		svg = svg + parms.getSVG();
		svg = svg + ">" + rtxt + "</tspan>";
		return svg;
	}

	public String getHTML(){
		if(isNull) return "";
		if(isBreak) return "<BR>";
		String rtxt = txt.replaceAll("<", "&lt;");
		rtxt = rtxt.replaceAll(">", "&gt;");
		String htmlall[] = parms.getHTML();
		String html = htmlall[0] + rtxt + htmlall[1];
		return html;
	}

	public void displaySel(int x, int y, float s, Graphics2D g2){
		if(sel > -1){
			int start = dx;
			if(sel > 0) start = adv[sel - 1];
			Point2D.Double first = new Point2D.Double(s * (x + start), s * (y + dy));
			Point2D.Double second = new Point2D.Double(s * (x + start), s * (y + dy - (s * parms.size)));
			Line2D.Double border = new Line2D.Double(first, second);
			float swidth = 0.5f;
			BasicStroke bs = new BasicStroke(swidth);
			g2.setStroke(bs);
			g2.setPaint(Color.gray);
			g2.draw(border);
		}			
	}

	public void initSel(){
		sel = -1;
		sel2 = -2;
	}

	public int findSel(int sel){
		if(sel < txt.length()){
			sel2 = -4;
			this.sel = sel;
			return -1;
		}
		sel2 = -3;
		int left = sel - txt.length();
		this.sel = txt.length();
		return (left);
	}

	public int findSel2(int sel){
		if(sel < txt.length()){
			if(sel2 != -4) this.sel = 0;
			sel2 = sel;
			return -1;
		}
		int left = sel - txt.length();
		if(sel2 == -2) {
			this.sel = 0;
			sel2 = txt.length();
		}
		else sel2 = txt.length(); 
		return (left);
	}

	public int hitSel(int x, int y){
		if(y < dy - parms.size) return 0;
		if(y > dy) return (-txt.length());
		for(int i = 0; i < txt.length(); i++){
			int mid = 0;
			if(i == 0) mid = ((2 * dx) + adv[0])/3;
			else mid = (2 * adv[i-1]+ (1 *adv[i]))/3;
			if(x < mid) return (i);
		}
		return (-txt.length());
	}

	public void drawStringX(Graphics2D g2, String txt0, int x0, int x, int y, int adv[], int i, float s){
		for(int j = 0; j < txt0.length(); j++){
			g2.drawString(txt0.substring(j, j+1), s * x, s * y);
			int a = x - x0;
			if((i + j) > 0) a = adv[i + j - 1];
			x = x + adv[i + j] - a;
		}
	}

	public void shadeSel(int x, int y, float s, Graphics2D g2){
		if(isBreak) return;
		if(!((sel2 <= sel) || (sel < 0))){
			int a = 0;
			if(sel > 0) a = adv[sel - 1];
			else a = dx;
			float fx = s*(x + a);
			float fy = s*(y + dy);
			float fdy = s*parms.size;
			float fdx = s*(adv[sel2 - 1] - a);
			float swidth = 0.5f;
			BasicStroke bs = new BasicStroke(swidth);
			g2.setStroke(bs);
			g2.setPaint(Color.gray);
			g2.fill(new Rectangle2D.Double(fx, fy - fdy, fdx, fdy));				
		}
	}

	public void display(int x, int y, float s, Graphics2D g2){
		if(isBreak) return;
		Font f = new Font(parms.font, parms.getStyle(), (int) (s * parms.size));
		g2.setFont(f);
		if((sel2 <= sel) || (sel < 0)){
			g2.setPaint(parms.color);
			//g2.drawString(txt, (int)(s*(x + dx)), (int)(s*(y + dy)));
			drawStringX(g2, txt, x, x + dx, y + dy, adv, 0, s);
		}
		else{
			String txt0 = "";
			if(sel > 0){
				txt0 = txt.substring(0, sel);
				g2.setPaint(parms.color);
				//g2.drawString(txt0, (int)(s*(x + dx)), (int)(s*(y + dy)));
				drawStringX(g2, txt0, x, x + dx, y + dy, adv, 0, s);	
			}	
			txt0 = txt.substring(sel, sel2); 
			int a = 0;
			if(sel > 0) a = adv[sel - 1];
			else a = dx;
			float fx = s*(x + a);
			float fy = s*(y + dy);
			//float fdy = s*parms.size;
			//float fdx = s*(adv[sel2 - 1] - a);
			//float swidth = 0.5f;
			//BasicStroke bs = new BasicStroke(swidth);
			//g2.setStroke(bs);
			//g2.setPaint(Color.gray);
			//g2.fill(new Rectangle2D.Double(fx, fy - fdy, fdx, fdy));
			g2.setPaint(Color.white);
			//g2.drawString(txt0, (int) fx, (int) fy);
			drawStringX(g2, txt0, x, x + a, y + dy, adv, sel, s);	
			if(txt.length() > sel2){
				txt0 = txt.substring(sel2, txt.length());
				g2.setPaint(parms.color);
				a = adv[sel2 - 1];
				//g2.drawString(txt0, (int)(s*(x + a)), (int)(s*(y + dy)));
				drawStringX(g2, txt0, x, x + a, y + dy, adv, sel2, s);
			}
						
		}
	}
}
